/**
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
package metridoc.component.sqlplus;

import java.util.Map;

import javax.sql.DataSource;

import org.apache.camel.Endpoint;
import org.apache.camel.impl.DefaultComponent;
import org.apache.camel.impl.ProcessorEndpoint;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class SqlPlusComponent extends DefaultComponent {
	
	private static final String FETCH_SIZE_MIN = "min";
	private static final Logger LOG = LoggerFactory.getLogger(SqlPlusComponent.class);

	@Override
	protected Endpoint createEndpoint(String uri, String remaining, Map<String, Object> parameters) throws Exception {
		String dataSourceKey = getAndRemoveParameter(parameters, "dataSource", String.class);
		DataSource dataSource = getCamelContext().getRegistry().lookup(dataSourceKey, DataSource.class);
		remaining = remaining.trim();
		if (remaining.endsWith(".sql")) {
			return createFileEndpoint(uri, remaining, parameters, dataSource);
		} else if (remaining.startsWith("select")) {
			return createStreamEndpoint(uri, remaining, parameters, dataSource);
		} else {
			return createUpdateEndpoint(uri, remaining, parameters, dataSourceKey);
		}
	}

	private Endpoint createStreamEndpoint(String uri, String query, Map<String, Object> parameters,
			DataSource dataSource) throws Exception {
		int fetchSize = getFetchSize(parameters);
		SqlStreamEndpoint endpoint = new SqlStreamEndpoint(uri, this, dataSource, query);
		endpoint.setFetchSize(fetchSize);
		return endpoint;
	}

	private Endpoint createFileEndpoint(String uri, String filePath, Map<String, Object> parameters,
			DataSource dataSource) throws Exception {
		return new ProcessorEndpoint(uri, getCamelContext(), new SqlFileRouteProcessor(filePath, dataSource));
	}

	private Endpoint createUpdateEndpoint(String uri, String query, Map<String, Object> parameters, String dataSourceKey)
			throws Exception {
		int batchSize = 0;
		if (parameters.containsKey("batchSize")) {
			batchSize = Integer.parseInt((String) parameters.remove("batchSize"));
		}
		boolean detailedOutput = Boolean.parseBoolean((String) parameters.remove("detailedOutput"));
		return new ProcessorEndpoint(uri, this, new SqlPlusUpdateProcessor(dataSourceKey, query, batchSize,
				detailedOutput));
	}

	private int getFetchSize(Map<String, Object> parameters) {
		int fetchSize = 0;
		String fetchSizeString = getAndRemoveParameter(parameters, "fetchSize", String.class);
		if (FETCH_SIZE_MIN.equals(fetchSizeString)) {
			fetchSize = Integer.MIN_VALUE;
		} else {
			try {
				fetchSize = Integer.parseInt(fetchSizeString);
			} catch (NumberFormatException e) {
				LOG.warn("Invalid fetch size: " + fetchSizeString + ".");
			}
		}
		LOG.debug("Fetch size is " + fetchSize);
		return fetchSize;
	}
}
