/**
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
package metridoc.component.sqlplus;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import javax.sql.DataSource;

import metridoc.impl.iterator.SqlIterator;
import metridoc.utils.Assert;
import org.apache.camel.Exchange;
import org.apache.camel.Message;
import org.apache.camel.Processor;
import org.apache.camel.spi.Registry;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;
import org.springframework.jdbc.core.namedparam.SqlParameterSourceUtils;
import org.springframework.jdbc.core.simple.SimpleJdbcTemplate;
import org.springframework.jdbc.support.rowset.SqlRowSet;

/**
 *
 * helps with batch updates.  The camel-sql processor only does single statements, this should help camel with speeding
 * up etl batch processing.  Using a database's loading tools is generally the fastest way to insert into a table
 * (by around 8 to 10 times faster), but this should be good enough for most scenarios.  Certain database loading utils
 * have pretty terrible feedback, so depending on the driver using this should provide better feedback when something
 * goes wrong.
 *
 * @author Tommy Barker
 */
public class SqlPlusUpdateProcessor implements Processor {

    private String dataSource;
    private SimpleJdbcTemplate simpleJdbcTemplate;
    SqlPlusUpdateEndpointProcessorUtils utils = new SqlPlusUpdateEndpointProcessorUtils();
    private String sql;
    private int batchSize;
    private boolean detailedOutput;

    public SqlPlusUpdateProcessor(String dataSource, String sql, int batchSize, boolean detailedOutput) {
    	this(dataSource, sql, batchSize);
    	this.detailedOutput = detailedOutput;
    }
    
    public SqlPlusUpdateProcessor(String dataSource, String sql, int batchSize) {
        Assert.notEmpty(dataSource, "The dataSource cannot be null or empty");
        Assert.notEmpty(sql, "sql cannot be null or empty");
        this.dataSource = dataSource;
        this.sql = utils.convertHashSql(sql);
        this.batchSize = batchSize;
    }

    @Override
    public void process(Exchange exchange) throws Exception {
    	Object updates = null;
    	if (simpleJdbcTemplate == null) {
            simpleJdbcTemplate = utils.getJdbcTemplate(dataSource, exchange);
        }
    	Message in = exchange.getIn();
    	SqlRowSet sqlRowSet = in.getBody(SqlRowSet.class);
    	if(sqlRowSet == null){
    		List body = in.getBody(List.class);
    		if(body == null){
    			Object object = in.getBody(Object.class);
    			if(object != null){
    				body = new ArrayList();
    				body.add(object);
    			}
    		}
    		Assert.notNull(body, "the body is either null or is not of expected type");
        
    		updates = utils.batchUpdate(body, simpleJdbcTemplate, sql, detailedOutput);
        }else{
        	Assert.isTrue(batchSize > 0, "The batchSize should be positive number");
        	updates = utils.batchUpdate(sqlRowSet, simpleJdbcTemplate, sql, batchSize, detailedOutput);
        }
        exchange.getOut().setHeader(SqlPlusConstants.UPDATES, updates);
    }

    static class SqlPlusUpdateEndpointProcessorUtils {

        SimpleJdbcTemplate getJdbcTemplate(String dataSourceRef, Exchange exchange) {
            Registry registry = exchange.getContext().getRegistry();

            DataSource dataSource = registry.lookup(dataSourceRef, DataSource.class);
            SimpleJdbcTemplate template = new SimpleJdbcTemplate(dataSource);

            return template;
        }

        String convertHashSql(String sql) {
            return sql.replaceAll("#", ":");
        }

        /**
         * performs the batch update
         * 
         * @param values the values to update
         * @param simpleJdbcTemplate template to perform the database operations
         * @param sql update sql
         * @return update count for each update ran (one update per value in values)
         */
        Object batchUpdate(List values, SimpleJdbcTemplate simpleJdbcTemplate, String sql, boolean detailedOutput) {
            SqlParameterSource[] batch;

            if (values.get(0) instanceof Map) {
                Map[] arrayForTransfer = new Map[values.size()];
                values.toArray(arrayForTransfer);
                batch = SqlParameterSourceUtils.createBatch(arrayForTransfer);
            } else {
                batch = SqlParameterSourceUtils.createBatch(values.toArray());
            }

            int[] updateCounts = simpleJdbcTemplate.batchUpdate(sql, batch);
            return detailedOutput?updateCounts:getSum(updateCounts);
        }
        
        /**
         * Performs the batch updates based on SqlRowSet content
         * @param sqlRowSet data source
         * @param simpleJdbcTemplate template to perform the database operations
         * @param sql update sql
         * @param batchSize size of single batch update input
         * @return update count for each update ran (one update per row in sqlRowSet)
         */
		Object batchUpdate(SqlRowSet sqlRowSet,
				SimpleJdbcTemplate simpleJdbcTemplate, String sql,
				int batchSize, boolean detailedOutput) {
			SqlIterator iterator = new SqlIterator(sqlRowSet);
			List<Map<String, Object>> values = new ArrayList<Map<String, Object>>();
			List<int[]> updateCountsList = new ArrayList<int[]>();
			int updateCountsFlatSize = 0;
			Object updateCounts = null;
			int updateCountsSum = 0;
			while (iterator.hasNext()) {
				if (values.size() == batchSize) {
					updateCounts = batchUpdate(values, simpleJdbcTemplate, sql, detailedOutput);
					if (detailedOutput) {
						updateCountsFlatSize += ((int[]) updateCounts).length;
						updateCountsList.add((int[]) updateCounts);
					} else {
						updateCountsSum += (Integer) updateCounts;
					}
					values.clear();
				}
				values.add((Map<String, Object>) iterator.next());
			}

			updateCounts = batchUpdate(values, simpleJdbcTemplate, sql, detailedOutput);
			if (detailedOutput) {
				updateCountsFlatSize += ((int[]) updateCounts).length;
				updateCountsList.add((int[]) updateCounts);
			} else {
				updateCountsSum += (Integer) updateCounts;
			}

			return detailedOutput ? toFlatArray(updateCountsList,
					updateCountsFlatSize) : updateCountsSum;
		}
        
        private int getSum(int[] array){
        	int result = 0;
        	for (int i = 0; i < array.length; i++) {
				result += array[i];
			}
        	return result;
        }
        
        private int[] toFlatArray(List<int[]> list, int flatArraySize){
        	int[] result = new int[flatArraySize];
        	int[] array; int k = 0;
        	for (int i = 0; i < list.size(); i++) {
        		array =list.get(i);
        		for (int j = 0; j < array.length; j++) {
        			result[k] = array[j];
        			k++;
				}
			}
        	return result;
        }
    }
}
